-- I HATE writing GUIs
DroidButtons.ui = {}
local db = DroidButtons
local ui = db.ui
ui.dialog = {}

local bool_map  = {[true] = 'ON', [false] = 'OFF'}
local state_map = {ON = true,  OFF = false}
local val_map   = {[true] = 1, [false] = 0}

local default = db.defaults

local recreate_default_touch_regions
local recreate_custom_touch_regions
local remove_touch_to_place_regions

-- Variables for GUI controls
local fontsize = 20
local buttonsize = '75x40'
local margin_vbox = '20x20'
local margin_buttonbox = '0x10' -- not a hex value.

local field_page_side_x = 340
local field_page_side_y = 160
local field_image_side = 140
local field_image_half = field_image_side / 2
local field_page_pad_x = (field_page_side_x - field_image_side) / 2
local field_page_pad_y = (field_page_side_y - field_image_side) / 2

commandlist, joystick_cmds, joycmdlist2, joycmdlist3, joycmdlist = dofile('vars.lua')
joycmdlist.font = fontsize
joycmdlist.size = 120

local type_radio_list = {'button', 'field'}
for i, v in ipairs(type_radio_list) do type_radio_list[v] = i end

-- localize functions and local functions
local rcreate = gkinterface.CreateTouchRegion
local rdestroy = gkinterface.DestroyTouchRegion
local tinsert = table.insert
local tremove = table.remove
local wInt = gkini.WriteInt
local wStr = gkini.WriteString

local function ispluscommand(str) 
	if (type(str) == 'string') then
		if (str:match('^+')) then return true end
	end
	return false
end

local badvalue = {
	create = function(self)
		local bv = {
			set = function(self, ctrl, state)
				-- pass the control and it's state
				if (not self[ctrl]) then self[ctrl] = #self + 1 end
				self[self[ctrl]] = state
			end,
			get = function(self)
				-- calculates the overall state based on the four fields
				local state = false
				for _, v in ipairs(self) do 
					state = v or state
					if (state) then break end
				end
				return state
			end
		}
		return bv
	end
}

local function number_var_check(ctrl, dflt, updt)
	-- check if ctrl.value has a number value and return that value
	-- if not a number then return the specified default value.
	-- set ctrl.value to the default if updt is true
	local val = tonumber(ctrl.value)
	if (not val) then
		val = dflt
		if (updt) then ctrl.value = val end
	end
	return val
end

local function save_options()
	wInt(db.major, 'DefaultWidth', tonumber(db.defaults.button_width))
	wInt(db.major, 'DefaultHeight', tonumber(db.defaults.button_height))
	wInt(db.major, 'DefaultFieldWidth', tonumber(db.defaults.field_width))
	wInt(db.major, 'DefaultFieldHeight', tonumber(db.defaults.field_height))
	wInt(db.major, 'DefaultFont', tonumber(db.defaults.font))
	wInt(db.major, 'SnapToGrid', tonumber(val_map[db.snap_to_grid]))
	wInt(db.major, 'GridSize', tonumber(db.grid_size))
	wInt(db.major, 'LongPressDelay', tonumber(db.long_press_delay))
	wInt(db.major, 'DisableDefaultUI', tonumber(val_map[db.disable_default_ui]))
end

local function list2(params)
	-- this function expects a table
	params.expand = params.expand or 'NO'
	params.list = params.list or {}
	params.size = params.size or 50
	params.value = params.value or ''
	params.active = params.active or 'YES'
	local textbox = iup.text{size = params.size, font = params.font, expand = params.expand, value = params.value, active = params.active}
	local dropdown = iup.list{dropdown = 'YES', size = params.size + 20, font = params.font, expand = params.expand, active = params.active, fgcolor = '0 0 0 0', visible_items = params.visible_items, action = function(self, t, i, v)
		textbox.value = t
		iup.SetFocus(textbox)
	end}
	for i, v in ipairs(params.list) do 
		dropdown[i] = v 
		if (v == params.value) then dropdown.value = i end
	end
	local zbox = iup.zbox{
		--dropdown must be before textbox within the zbox to create the proper z-order
		iup.hbox{dropdown,},
		iup.hbox{textbox, iup.fill{size = 20}},
		all = 'YES',
		alignment = 'NW',
	}
	return zbox, textbox, dropdown
end

-- -----------------------------------------------------------

local function create_edit_dialog()
	local buttonindex -- this will need to be set by dialog:populate(index)
	local dirtyrecord = false
	local newbutton = false
	local default = db.defaults

	local title
	-- Button reion description controls
	local cmd1, cmd1text, cmd1list, cmd2, cmd2text, cmd2list, cmd2c, cmd2ctext, cmd2clist
	-- Field region description controls
	local fieldregionpic, xaxiscmd, yaxiscmd, xinvert, yinvert, xmouselook, ymouselook, relative, nonlinear
	-- definition controls
	local xpos, ypos, width, height, font

	local place
	local type_radio, definition_switch
	local savechanges, close

	local badvalue = badvalue:create() -- Used for tracking bad values in the four positioning fields

	local function make_dialog_dirty()
		-- used as the action callback for a number of textboxes
		-- If a value has changed, mark the form as dirty and allow saving
		dirtyrecord = true
		if (not badvalue:get()) then
			savechanges.active = 'YES'
		end
		return iup.DEFAULT
	end

	local function verify_number_entry(self, c, after)
		-- used as the action callback for a number of textboxes
		-- set badvalue for the current field based on entry
		if (after ~= '') then
			local num = tonumber(after)
			if (num and num >= 0) then
				self.fgcolor = '255 255 255'
				badvalue:set(self, false)
			else
				self.fgcolor = '255 0 0'
				badvalue:set(self, true)
			end
		end

		-- verify if any of the fields have a badvalue
		if (badvalue:get()) then savechanges.active = 'NO' else savechanges.active = 'YES' end

		return make_dialog_dirty()
	end
	
	-- Button region description controls.  They will need to be populated later.
	title = iup.text{size = 170, font = fontsize}
	cmd1, cmd1text, cmd1list = list2{size = 150, font = fontsize, list = commandlist, visible_items = 10}
	cmd2, cmd2text, cmd2list = list2{size = 150, font = fontsize, list = commandlist, visible_items = 10}
	cmd2c, cmd2ctext, cmd2clist = list2{size = 150, font = fontsize, list = commandlist, visible_items = 10}
	
	-- Field region description controls
	fieldregionpic = iup.label{cx = field_page_pad_x, cy = field_page_pad_y, size = field_image_side..'x'..field_image_side, title = '',image = IMAGE_DIR..'hud_radar.png'}
	xaxiscmd = iup.stationsublist(joycmdlist); xaxiscmd.cx = (field_page_pad_x - 95); xaxiscmd.cy = ((field_page_pad_y + field_image_half) - 10); xaxiscmd.action = make_dialog_dirty
	yaxiscmd = iup.stationsublist(joycmdlist); yaxiscmd.cx = (field_page_pad_x - 42); yaxiscmd.cy = field_page_pad_y + 5; yaxiscmd.action = make_dialog_dirty
	xinvert = iup.stationtoggle{cx = ((field_page_side_x - field_page_pad_x) - 20), cy = ((field_page_pad_y + field_image_half) - 10), title = 'Invert X Axis', font = fontsize, action = make_dialog_dirty}
	yinvert = iup.stationtoggle{cx = ((field_page_pad_x + field_image_half) - 8), cy = ((field_page_side_y - field_page_pad_y) - 23), title = 'Invert Y Axis', font = fontsize, action = make_dialog_dirty}
	xmouselook = iup.stationtoggle{cx = ((field_page_side_x - field_page_pad_x) - 20), cy = ((field_page_pad_y + field_image_half) - 10 + 20), title = 'Mouselook X', font = fontsize, action = make_dialog_dirty}
	ymouselook = iup.stationtoggle{cx = ((field_page_pad_x + field_image_half) - 8), cy = ((field_page_side_y - field_page_pad_y) - 23 + 20), title = 'Mouselook Y', font = fontsize, action = make_dialog_dirty}
	relative = iup.stationtoggle{title = 'Relative', font = fontsize, action = make_dialog_dirty}
	nonlinear = iup.stationtoggle{title = 'Non-Linear', font = fontsize, action = make_dialog_dirty}

	-- definition controls
	xpos = iup.text{size = 64, font = fontsize, action = verify_number_entry}
	ypos = iup.text{size = 64, font = fontsize, action = verify_number_entry}
	width = iup.text{size = 64, font = fontsize, action = verify_number_entry}
	height = iup.text{size = 64, font = fontsize, action = verify_number_entry}
	font = iup.text{size = 64, font = fontsize, action = verify_number_entry}
	
	place = iup.stationbutton{title = 'Place', size = buttonsize, font = fontsize}

	definition_switch = iup.stationbutton{title = 'Advanced', size = 64}
	
	savechanges = iup.stationbutton{title = 'Save', size = buttonsize, font = fontsize, active = 'NO'}
	close = iup.stationbutton{title = 'Close', size = buttonsize, font = fontsize}

	-- -----------------------------------------------------------

	title.action = make_dialog_dirty
	cmd1text.action = make_dialog_dirty
	cmd2ctext.action = make_dialog_dirty

	local function toggle_cmd2c(t)
		if (ispluscommand(t)) then 
			cmd2ctext.active = 'YES'
			cmd2clist.active = 'YES'
		else
			cmd2ctext.value = ''
			cmd2ctext.active = 'NO'
			cmd2clist.value = 0
			cmd2clist.active = 'NO'
		end
	end

	function cmd2text:action(c, after)
		toggle_cmd2c(after)
		return make_dialog_dirty()
	end

	local old_cmd1list_action = cmd1list.action
	function cmd1list:action(t, i, v)
		old_cmd1list_action(self, t, i, v)
		return make_dialog_dirty()
	end

	local old_cmd2list_action = cmd2list.action
	function cmd2list:action(t, i, v)
		old_cmd2list_action(self, t, i, v)
		toggle_cmd2c(t)
		return make_dialog_dirty()
	end

	local old_cmd2clist_action = cmd2clist.action
	function cmd2clist:action(t, i, v)
		old_cmd2clist_action(self, t, i, v)
		return make_dialog_dirty()
	end

	-- vvv BEGIN SUB GUI vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
	local function show_placement_interface(dialog)
		-- dialog = the dialog that this function should re-open when it's done./
		local Xres, Yres = db.xres, db.yres
		db.place_button_mode = true

		if (remove_touch_to_place_regions) then
			db:destroy_touch_to_place_regions()
		end

		HUD:DestroyAimTouchRegion()
		HUD:DestroyTouchLayer()
		recreate_default_touch_regions = true
		db:destroy_button_regions()
		db:destroy_field_regions()
		recreate_custom_touch_regions = true

		HUD.help_text.title = 'Touch the screen to place the button.'
		
		local save = iup.stationbutton{title = 'Save', size = '64x64', font = fontsize, cx = 0, cy = 0}
		local cancel = iup.stationbutton{title = 'Cancel', size = '64x64', font = fontsize, cx = 70, cy = 0}

		iup.Append(db.cbox, save)
		iup.Append(db.cbox, cancel)
		iup.Refresh(db.cbox)

		local save_region = gkinterface.CreateTouchRegion(nil, nil, nil, false, false, false, false, false, false, tonumber(save.cx), tonumber(save.cy), tonumber(save.cx) + 63, tonumber(save.cy) + 63)
		local cancel_region = gkinterface.CreateTouchRegion(nil, nil, nil, false, false, false, false, false, false, tonumber(cancel.cx), tonumber(cancel.cy), tonumber(cancel.cx) + 63, tonumber(cancel.cy) + 63)
		local screen_region = gkinterface.CreateTouchRegion(nil, nil, nil, false, false, false, false, false, false, 0, 0, Xres, Yres, 1)

		db.regions[screen_region] = function(pW, pH) 
			--local default = db.defaults
			--create a copy of the current buttondef so we aren't modifying saved values
			local b_type = type_radio.value.title:lower()
			local buttondef = {
				title = title.value or '',
				type = b_type,
				x = math.floor(pW * Xres), -- TOUCH_RELEASED now returns percent of width and height. Not sure how always rounding down will affect things.
				y = math.floor(pH * Yres),
				w = number_var_check(width, default[b_type..'_width'], false),
				h = number_var_check(height, default[b_type..'_height'], false),
				font = number_var_check(font, default.font, false),
			}

			for axis, leg in pairs({x = 'w', y = 'h'}) do
				-- adjust axis by half the leg to center the button
				buttondef[axis] = buttondef[axis] - math.floor(buttondef[leg]/2)

				-- bring axis back within the screen if necessary
				if (buttondef[axis] < 0) then 
					buttondef[axis] = 0
				elseif (buttondef[axis] + buttondef[leg] - 1 > db[axis..'res']) then 
					buttondef[axis] = db[axis..'res'] - buttondef[leg] + 1
				end
				
				-- snap to grid if applicable.  I chose to always snap towards 0.
				if (db.snap_to_grid) then
					local r = buttondef[axis] % db.grid_size
					buttondef[axis] = buttondef[axis] - r
				end
			end

			if (newbutton and not db.buttons[buttonindex]) then 
				db:add_button(buttondef)
				iup.Append(db.cbox, db.buttons[buttonindex])
				db.cbox:map()
				iup.Refresh(db.cbox)
			else
				db:update_button(buttonindex, buttondef)
			end
		end

		db.regions[cancel_region] = function(saved)
			if (type(saved) ~= 'boolean') then saved = false end
			db.place_button_mode = false

			if (not saved) then
				if (newbutton) then
					if(db.buttons[buttonindex]) then
						iup.Detach(db.buttons[buttonindex])
						iup.Destroy(db.buttons[buttonindex])
						db.buttons[buttonindex] = nil
					end
				else
					db:update_button(buttonindex) --put the button back where it was
					--iup.Refresh(db.cbox)
					xpos.value = db.buttondefs[buttonindex].x
					ypos.value = db.buttondefs[buttonindex].y
				end
			end

			HUD.help_text.title = ''
			--rdestroy(screen_region)
			--db.regions[screen_region] = nil

			iup.Detach(save);   iup.Destroy(save);   save = nil;   --rdestroy(save_region)
			iup.Detach(cancel); iup.Destroy(cancel); cancel = nil; --rdestroy(cancel_region)

			ShowDialog(dialog)
		end

		db.regions[save_region] = function()
			xpos.value = db.buttons[buttonindex].x
			ypos.value = db.buttons[buttonindex].y
			make_dialog_dirty()
			db.regions[cancel_region](true)
		end

		remove_touch_to_place_regions = true

		HideDialog(dialog)
	end
	-- ^^^ END SUB GUI ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	                     
	function place:action()
		show_placement_interface(iup.GetDialog(self))
	end

	function savechanges:action()
		local default = db.defaults
		local buttondef = db.buttondefs[buttonindex]

		buttondef.type = type_radio.value.title:lower()
		buttondef.title = title.value or ''
		buttondef.x = number_var_check(xpos, 0, true)
		buttondef.y = number_var_check(ypos, 0, true)
		buttondef.w = number_var_check(width, default[buttondef.type..'_width'], true)
		buttondef.h = number_var_check(height, default[buttondef.type..'_height'], true)
		buttondef.font = number_var_check(font, default.font, true)
		if (buttondef.type == 'button') then
			buttondef.command1 = cmd1text.value; if (buttondef.command1 == '') then buttondef.command1 = nil end
			buttondef.command2 = cmd2text.value; if (buttondef.command2 == '') then buttondef.command2 = nil end
			buttondef.command2cancel = cmd2ctext.value; if (buttondef.command2cancel == '') then buttondef.command2cancel = nil end
		elseif (buttondef.type == 'field') then
			--test.varprint(' / ', 'xaxiscmd', xaxiscmd.value, 'xjoycmd', joycmdlist3[xaxiscmd.value], 'yaxiscmd', yaxiscmd.value, 'yjoycmd', joycmdlist3[yaxiscmd.value])
			buttondef.command1 = joycmdlist3[tonumber(xaxiscmd.value)]--; if (buttondef.command1 == 'NONE') then buttondef.command1 = nil end
			buttondef.command2 = joycmdlist3[tonumber(yaxiscmd.value)]--; if (buttondef.command2 == 'NONE') then buttondef.command2 = nil end
			buttondef.invertX = state_map[xinvert.value]
			buttondef.invertY = state_map[yinvert.value]
			buttondef.mouselookX = state_map[xmouselook.value]
			buttondef.mouselookY = state_map[ymouselook.value]
			buttondef.relative = state_map[relative.value]
			buttondef.nonlinear = state_map[nonlinear.value]
			--test.print(test.stringify(buttondef, nil, true))
		end
		
		if (newbutton) then
			type_radio[1][(type_radio_list[type_radio.value.title:lower()] % 2) +1].active = 'NO'
			if (not db.buttons[buttonindex]) then 
				db:add_button(buttondef)
				iup.Append(db.cbox, db.buttons[buttonindex])
				db.cbox:map()
			else
				db:update_button(buttonindex)
			end
		else
			db:update_button(buttonindex)
		end

		db:save_buttondefs()
		ui.buttonlist:populate()
		savechanges.active = 'NO'
		dirtyrecord = false
		newbutton = false
		recreate_custom_touch_regions = true
	end

	function close:action()
		if (newbutton) then 
			if(db.buttons[buttonindex]) then
				iup.Detach(db.buttons[buttonindex])
				iup.Destroy(db.buttons[buttonindex])
				db.buttons[buttonindex] = nil
			end
			db.buttondefs[buttonindex] = nil
			newbutton = false
		end

		if (dirtyrecord) then
			db:update_button(buttonindex) --put the button back where it was
			savechanges.active = 'NO'
			dirtyrecord = false
		end

		ShowDialog(ui.dialog, iup.CENTER, iup.CENTER)
		HideDialog(iup.GetDialog(self))
	end

	-- -----------------------------------------------------------

	-- Description Tab -------------------------------------------
	local description_zbox_items = {alignment = 'ATOP', bgcolor = "255 255 200 100 *"}
	description_zbox_items[1] = iup.vbox{
		iup.fill{size = 10},
		iup.hbox{iup.label{title = 'Short-press: ', font = fontsize, expand = 'YES'}, cmd1},
		iup.hbox{iup.label{title = 'Long-press: ', font = fontsize, expand = 'YES'}, cmd2},
		iup.hbox{iup.label{title = 'Long-press cancel: ', font = fontsize, expand = 'YES'}, cmd2c},
		gap=10,
		alignment = 'ATOP'
	}
	description_zbox_items[2] = iup.vbox{
		iup.cbox{
			fieldregionpic,
			xaxiscmd,
			yaxiscmd,
			xinvert,
			yinvert,
			xmouselook,
			ymouselook,
			size = field_page_side_x..'x'..field_page_side_y,
			active = "YES"
		},
		iup.fill{size = 10},
		iup.hbox{
			iup.fill{},
			relative,
			nonlinear,
			iup.fill{},
			gap = 32,
		},
	}

	local description_zbox = iup.zbox(description_zbox_items)
	
	local description = iup.vbox{
		iup.hbox{iup.label{title = 'Title: ', font = fontsize, expand = 'YES'}, title},
		description_zbox,
		gap=10,
		margin = margin_vbox,
	}

	local type_radio_items = {alignment = 'ACENTER',gap = 15,}
	type_radio_items[1] = iup.stationradio{
		title = 'Button',
		font = fontsize,
		action = function(self, v) 
			description_zbox.value = description_zbox_items[1]
			if (type_radio_items[2].active == 'YES') then
				width.value = default[self.title:lower()..'_width']
				height.value = default[self.title:lower()..'_height']
			end
		end,
	}
	type_radio_items[2] = iup.stationradio{
		title = 'Field',
		font = fontsize,
		action = function(self, v)
			description_zbox.value = description_zbox_items[2]
			if (type_radio_items[1].active == 'YES') then
				width.value = default[self.title:lower()..'_width']
				height.value = default[self.title:lower()..'_height']
			end
		end,
	}
	--test.type_radio_items = type_radio_items

	type_radio = iup.radio{
		iup.hbox(type_radio_items),
		set_value = function(self, v)
			self.value = type_radio_items[v]
			type_radio_items[v]:action()
		end,
	}
	--test.type_radio = type_radio

	local description_tab =	iup.vbox{
		iup.pdasubframe_nomargin{iup.hbox{type_radio, iup.fill{expand = 'YES'}, margin = '5x5'}},
		iup.pdasubframe_nomargin{description,},
		gap = 2,
		--margin = margin_vbox,
		tabtitle = "Description",
	}

	-- Definition Tab --------------------------------------------
	local definition_zbox_items = {alignment = 'ACENTER', bgcolor = "255 255 200 100 *"}
	definition_zbox_items[1] = iup.vbox{
		iup.hbox{iup.fill{}, place, iup.fill{}},
		gap = 10,
		margin = margin_buttonbox,
	}
	definition_zbox_items[2] = iup.vbox{
		iup.hbox{iup.fill{}, iup.label{title = 'X pos: ', font = fontsize}, xpos, iup.fill{expand = 'YES'}, iup.label{title = 'Y pos: ', font = fontsize}, ypos, iup.fill{}},
		iup.hbox{iup.fill{}, iup.label{title = 'Width: ', font = fontsize}, width, iup.fill{expand = 'YES'}, iup.label{title = 'Height: ', font = fontsize}, height, iup.fill{}},
		iup.hbox{iup.fill{}, iup.label{title = 'Font Size: ', font = fontsize}, font, iup.fill{}},
		iup.fill{expand = 'YES'},
		gap = 20,
		margin = margin_buttonbox,
	}

	local definition_zbox = iup.zbox(definition_zbox_items)
	definition_zbox.value = definition_zbox_items[1]
		
	function definition_switch:action(mode)
		mode = mode or self.title
		if (mode == 'Basic') then
			self.title = 'Advanced'
			definition_zbox.value = definition_zbox_items[1]
		elseif (mode == 'Advanced') then
			self.title = 'Basic'
			definition_zbox.value = definition_zbox_items[2]
		end
	end

	local definition_tab = iup.vbox{
		iup.hbox{iup.fill{}, definition_switch},
		iup.pdasubframe_nomargin{definition_zbox,},
		gap = 2,
		tabtitle = "Definition",
	}

	local tabs = iup.pda_root_tabs{
		description_tab,
		definition_tab,
	}

	local main = iup.pdarootframe{
--		iup.pdarootframebg{
			iup.vbox{
				tabs,
				iup.hbox{iup.fill{}, savechanges, close, iup.fill{}, gap = 15, alignment = 'ACENTER'},
				gap=15,
				margin = margin_buttonbox,
			},
--		},
	}

	local dialog = iup.dialog{
		iup.vbox{
			iup.fill{},
			iup.hbox{
				iup.fill{},
				main,
				iup.fill{},
			},
			iup.fill{},
		},
		defaultesc=close,
		bgcolor="0 0 0 *",
		-- fullscreen="YES",
		title = 'DroidButtons v'..db.version..': Button Edit',
		border="NO",
		resize="NO",
		menubox="YES",
		topmost="YES",
	}

	function dialog:show_cb()
		tabs:SetTab(2)
	end
	
	function dialog:populate(index)
		if (not db.buttondefs[index]) then 
			db.buttondefs[index] = {type = 'button'} 
			newbutton = true
		end
		local default = db.defaults
		local buttondef = db.buttondefs[index]
		
		type_radio:set_value(type_radio_list[buttondef.type])
		title.value = tostring(buttondef.title or '')
		xpos.value = tostring(buttondef.x or '')
		ypos.value = tostring(buttondef.y or '')
		width.value = tostring(buttondef.w or default[buttondef.type..'_width'])
		height.value = tostring(buttondef.h or default[buttondef.type..'_height'])
		font.value = tostring(buttondef.font or default.font)

		if (buttondef.type == 'button') then
			cmd1text.value = tostring(buttondef.command1 or '')
			cmd2text.value = tostring(buttondef.command2 or '')
			cmd2ctext.value = tostring(buttondef.command2cancel or '')
			cmd2ctext.active = buttondef.command2cancel and 'YES' or 'NO'
			cmd2clist.active = buttondef.command2cancel and 'YES' or 'NO'

		elseif (buttondef.type == 'field') then
			xaxiscmd.value = joycmdlist2[(buttondef.command1 or 'NONE')]
			yaxiscmd.value = joycmdlist2[(buttondef.command2 or 'NONE')]
			xinvert.value = buttondef.invertX and 'ON' or 'OFF'
			yinvert.value = buttondef.invertY and 'ON' or 'OFF'
			xmouselook.value = buttondef.mouselookX and 'ON' or 'OFF'
			ymouselook.value = buttondef.mouselookY and 'ON' or 'OFF'
			relative.value = buttondef.relative and 'ON' or 'OFF'
			nonlinear.value = buttondef.nonlinear and 'ON' or 'OFF'

		else
			return false
		end

		if (not newbutton) then
			local type_idx = type_radio_list[buttondef.type]
			type_radio_items[type_idx].active = 'YES'
			type_radio_items[(type_idx%2)+1].active = 'NO'
		else
			type_radio_items[1].active = 'YES'
			type_radio_items[2].active = 'YES'
		end

		definition_switch:action('Basic')

		buttonindex = index
		return true
	end

	dialog:map()
	
	return dialog
	
end

-- -----------------------------------------------------------

local function create_options_dialog()
	local default = db.defaults
	local remove_default_ui_dialog_open = false

	local button_width = iup.text{size = 64, font = fontsize} 
	local button_height = iup.text{size = 64, font = fontsize}

	local field_width = iup.text{size = 64, font = fontsize} 
	local field_height = iup.text{size = 64, font = fontsize}
	
	local font = iup.text{size = 64, font = fontsize}
	local delay = iup.text{size = 64, font = fontsize}
	
	local snaptogrid = iup.stationtoggle{title = 'Snap-to-Grid', font = fontsize}
	local gridsize = iup.text{size = 64, font = fontsize}
	
	local lockimage = iup.label{title = '', image = db.IMAGE_DIR..'Lock.png', font = fontsize, size = '24x24', bgcolor = '255 255 255 255 *'}
	local lockbutton = iup.stationbutton{title = '', font = fontsize, size = '32x32', bgcolor = '255 255 255 75 +'}
	local remove_default_ui = iup.stationtoggle{title = 'Remove Default Touch Interface', font = fontsize, active = 'NO'}
	
	local save = iup.stationbutton{title = 'Save', size = buttonsize, font = fontsize}
	local cancel = iup.stationbutton{title = 'Cancel', size = buttonsize, font = fontsize}
	
	local unlock_timer = Timer()
	function lockbutton:action()
		if (lockimage.image == db.IMAGE_DIR..'Lock.png') then
			lockimage.image = db.IMAGE_DIR..'Open_lock.png'
			remove_default_ui.active = 'YES'
			unlock_timer:SetTimeout(5000, function() if (not remove_default_ui_dialog_open) then lockbutton:action() end end)
		else
			lockimage.image = db.IMAGE_DIR..'Lock.png'
			remove_default_ui.value = 'OFF'
			remove_default_ui.active = 'NO'
			unlock_timer:Kill()
		end
	end

	local remove_default_dlg_ok = function(self)
		unlock_timer:Kill()
		lockbutton.active = 'NO'
		lockimage.image = db.IMAGE_DIR..'Open_lock_disabled.png'
		remove_default_ui_dialog_open = false
		if (self) then HideDialog(iup.GetDialog(self)) end
	end

	local remove_default_ui_confirmation_dialog = msgdlgtemplate2(
		"By selecting this, you are choosing to remove the default UI and create your own.\n\n\127ff0000Do not contact Guild Software regarding any problems experienced with this feature enabled or this plugin installed.\n\n\127ffffffDo you wish to proceed?",
		'OK',
		remove_default_dlg_ok,
		'Cancel',
		function(self)
			unlock_timer:Kill()
			lockbutton:action()
			remove_default_ui_dialog_open = false
			HideDialog(iup.GetDialog(self))
		end
	)--, dlgbgcolor, alignment)

	local use_default_dlg_ok = function(self)
		lockbutton.active = 'YES'
		lockbutton:action()
		if (self) then HideDialog(iup.GetDialog(self)) end
	end

	local use_default_ui_confirmation_dialog = msgdlgtemplate2(
		"In order to re-enable to default touch interface, it will be necessary to reload the interface.\n\nIf you click OK, this will be done automatically for you upon clicking 'Save'.\n\n\127ffffffDo you wish to proceed?",
		'OK',
		use_default_dlg_ok,
		'Cancel',
		function(self)
			remove_default_ui.value = 'ON'
			HideDialog(iup.GetDialog(self))
		end
	)--, dlgbgcolor, alignment)

	function remove_default_ui:action(i)
		if (i == 1) then
			remove_default_ui_dialog_open = true
			ShowDialog(remove_default_ui_confirmation_dialog, iup.CENTER, iup.CENTER)
		else
			if (state_map[remove_default_ui.value] ~= db.disable_default_ui) then
				ShowDialog(use_default_ui_confirmation_dialog, iup.CENTER, iup.CENTER)
			else
				use_default_dlg_ok()
			end
		end
	end
	
	function cancel:action()
		ShowDialog(ui.dialog, iup.CENTER, iup.CENTER)
		HideDialog(iup.GetDialog(self))
	end
	
	function save:action()
		local new_disable_ui_val = state_map[remove_default_ui.value]
		local reload_after_save = false
		if (new_disable_ui_val ~= db.disable_default_ui) then
			if (new_disable_ui_val) then
				db:remove_default_ui()
			else
				reload_after_save = true
			end
		end
		default.button_width = tonumber(button_width.value)
		default.button_height = tonumber(button_height.value)
		default.field_width = tonumber(field_width.value)
		default.field_height = tonumber(field_height.value)
		default.font = tonumber(font.value)
		db.disable_default_ui = new_disable_ui_val
		db.long_press_delay = tonumber(delay.value)
		db.snap_to_grid = state_map[snaptogrid.value]
		db.grid_size = tonumber(gridsize.value)
		save_options()
		if (reload_after_save) then ReloadInterface() end
		cancel:action()
	end

	local function subdlg(ctrl)
		local dlg = iup.dialog{
			ctrl,
			border="NO",
			menubox="NO",
			resize="NO",
			bgcolor="0 0 0 0 *",
		}
		return dlg
	end

	local list_content = iup.hbox{
		iup.fill{expand = 'YES'},
		iup.vbox{
			iup.vbox{
				iup.label{title = 'Button', font = fontsize + 2},
				iup.hbox{
					iup.fill{size = 25},
					iup.vbox{
						iup.hbox{iup.label{title = 'Width: ', font = fontsize, expand = 'YES'}, button_width},
						iup.hbox{iup.label{title = 'Height: ', font = fontsize, expand = 'YES'}, button_height},
						gap = 5,
					},
				},
			},
			iup.vbox{
				iup.label{title = 'Field', font = fontsize + 2},
				iup.hbox{
					iup.fill{size = 25},
					iup.vbox{
						iup.hbox{iup.label{title = 'Width: ', font = fontsize, expand = 'YES'}, field_width},
						iup.hbox{iup.label{title = 'Height: ', font = fontsize, expand = 'YES'}, field_height},
						gap =5,
					},
				},
			},
			iup.hbox{iup.label{title = 'Font Size: ', font = fontsize, expand = 'YES'}, font},
			iup.hbox{iup.label{title = 'Long-press delay: ', font = fontsize, expand = 'YES'}, delay, iup.label{title = 'ms', font = fontsize}},
			iup.vbox{
				iup.hbox{snaptogrid},
				iup.hbox{iup.label{title = 'Grid Size: ', font = fontsize, expand = 'YES'}, gridsize},
				gap = 5,
			},
			gap = 20, 
		},
		iup.fill{expand = 'YES'},
		size = 'THIRD',
		margin = '20x5'
	}

	local listbox = iup.stationsublist{{}, control = "YES", size = 'THIRDxTHIRD', expand = "YES",}
	iup.Append(listbox, subdlg(list_content))

	local basic_tab = iup.vbox{
		listbox,
		tabtitle = 'Basic',
	}

	local adv_tab = iup.vbox{
		iup.fill{},
		iup.hbox{
			iup.fill{}, 
			iup.zbox{
				iup.vbox{
					iup.fill{size = 4},
					iup.hbox{iup.fill{size = 4},lockimage,iup.fill{size = 4}},
					iup.fill{size = 4},
				},
				lockbutton,
				all = 'YES',
			},
			remove_default_ui, 
			iup.fill{}, 
			gap = 20,
		},
		iup.fill{},
		tabtitle = 'Advanced',
	}

	local tabs = iup.pda_root_tabs{
		basic_tab,
		adv_tab,
	}

	local main = iup.pdarootframe{
--		iup.pdarootframebg{
			iup.vbox{
				tabs,
				iup.hbox{iup.fill{}, save, cancel, iup.fill{}, gap = 15, alignment = 'ACENTER'},
				gap=20,
				margin = margin_buttonbox
			},
--		},
	}

	local dialog = iup.dialog{
		iup.vbox{
			iup.fill{},
			iup.hbox{
				iup.fill{},
				main,
				iup.fill{},
			},
			iup.fill{},
		},
		defaultesc=cancel,
		bgcolor="0 0 0 *",
		-- fullscreen="YES",
		title = 'DroidButtons v'..db.version..': Options',
		border="NO",
		resize="NO",
		menubox="YES",
		topmost="YES",
	}

	function dialog:show_cb()
		button_width.value = default.button_width
		button_height.value = default.button_height
		field_width.value = default.field_width
		field_height.value = default.field_height
		font.value = default.font
		delay.value = db.long_press_delay
		snaptogrid.value = bool_map[db.snap_to_grid]
		gridsize.value = db.grid_size
		remove_default_ui.value = bool_map[db.disable_default_ui]
		tabs:SetTab(2)
		if (Subplatform == "XperiaPlay") then
			lockbutton.active = 'NO'
			lockimage.image = db.IMAGE_DIR..'Lock_disabled.png'
		elseif (db.disable_default_ui) then 
			remove_default_ui.active = 'YES'
			remove_default_dlg_ok()
		end
	end
	
	dialog:map()
	
	return dialog
end

function ui:build_dialog()
	local editdialog = create_edit_dialog()
	local optionsdialog = create_options_dialog()

	local dirtyrecord = false

	self.buttonlist = iup.stationsublist{size = '200x200', visible_items = 10, font = fontsize}

	self.deletebutton = iup.stationbutton{title = 'Delete', size = buttonsize, active = 'NO', font = fontsize}
	self.editbutton = iup.stationbutton{title = 'Edit', size = buttonsize, active = 'NO', font = fontsize}
	self.addbutton = iup.stationbutton{title = 'Add', size = buttonsize, active = 'YES', font = fontsize}

	self.optionsbutton = iup.stationbutton{title = 'Options', size = buttonsize, active = 'YES', font = fontsize}
	self.closebutton = iup.stationbutton{title = 'Close', size = buttonsize, active = 'YES', font = fontsize}

	function self.buttonlist:action(t, i, v)
		if     (v == 1) then
			ui.deletebutton.active = 'YES'
			ui.editbutton.active = 'YES'
		elseif (v == 0) then
			ui.deletebutton.active = 'NO'
			ui.editbutton.active = 'NO'
		end
	end
	
	function self.buttonlist:populate()
		for i, v in ipairs(db.buttondefs) do 
			local name = v.title
			if (name == '') then name = v.command1 or '' end
			self[i] = name..'('..v.type..')'
		end
		self[#db.buttondefs + 1] = nil
	end

	function self.deletebutton:action()
		local index = tonumber(ui.buttonlist.value)
		if (index < 1) then return end

		iup.Detach(db.buttons[index])
		iup.Destroy(db.buttons[index])
		tremove(db.buttons, index)
		tremove(db.buttondefs, index)
		db:save_buttondefs()

		if (HUD.IsVisible == true) then
			for i, v in pairs(db.regions) do 
				if (v == index) then 
					gkinterface.DestroyTouchRegion(i)
					db.regions[i] = nil 
					break 
				end 
			end
		end
		
		ui.buttonlist:populate()
		if (tonumber(ui.buttonlist.value) == 0) then
			ui.editbutton.active = 'NO'
			self.active = 'NO'
		end
	end

	function self.editbutton:action()
		local index = tonumber(ui.buttonlist.value)
		if (index < 1) then return end
		if (editdialog:populate(index)) then
			ShowDialog(editdialog, iup.CENTER, iup.CENTER)
			HideDialog(iup.GetDialog(self))
		end
	end
	
	function self.addbutton:action()
		if (editdialog:populate(#db.buttondefs + 1)) then
			ShowDialog(editdialog, iup.CENTER, iup.CENTER)
			HideDialog(iup.GetDialog(self))
		end
	end
	
	function self.optionsbutton:action()
		ShowDialog(optionsdialog, iup.CENTER, iup.CENTER)
		HideDialog(iup.GetDialog(self))
	end
	
	function self.closebutton:action()
		if (remove_touch_to_place_regions) then
			db:destroy_touch_to_place_regions()
			remove_touch_to_place_regions = false
		end

		if (recreate_default_touch_regions) then
			HUD:CreateAimTouchRegion()
			HUD:CreateTouchLayer()
			recreate_default_touch_regions = false
		end

		if (recreate_custom_touch_regions) then
			db:create_button_regions()
			db:create_field_regions()
			recreate_custom_touch_regions = false
		end

		HideDialog(iup.GetDialog(self))
	end
	
	local buttons_tab = iup.vbox{
		iup.fill{size = 5},
		iup.hbox{iup.fill{}, self.buttonlist, iup.fill{}},
		iup.hbox{iup.fill{}, self.addbutton, self.editbutton, self.deletebutton, iup.fill{}, gap = 15, margin = margin_buttonbox},
		tabtitle = 'Buttons',
	}


	local xaxiscmd, yaxiscmd, custom_accel, set_tilt
	local function make_dialog_dirty()
		-- used as the action callback for a number of textboxes
		-- If a value has changed, mark the form as dirty and allow saving
		dirtyrecord = true
		set_tilt.active = 'YES'

		return iup.DEFAULT
	end
	
	xaxiscmd = iup.stationsublist(joycmdlist); xaxiscmd.font = fontsize; 
	function xaxiscmd:action(t, i, v)
		if (v ~= 0) then
			if (i == tonumber(self.cursel)) then return else self.cursel = i end
			return make_dialog_dirty()
		end
	end
	yaxiscmd = iup.stationsublist(joycmdlist); yaxiscmd.font = fontsize; yaxiscmd.action = make_dialog_dirty
	function yaxiscmd:action(t, i, v)
		if (v ~= 0) then
			if (i == tonumber(self.cursel)) then return else self.cursel = i end
			return make_dialog_dirty()
		end
	end
	custom_accel = iup.stationtoggle{title = 'Use Custom Accelerometer Bindings', font = fontsize, action = make_dialog_dirty}
	
	set_tilt = iup.stationbutton{title = 'Set', font = fontsize, action = function(self)
		db.use_custom_accel = state_map[custom_accel.value]
		wInt(db.major, 'CustomAccel', val_map[db.use_custom_accel])
		db.accelx = joycmdlist3[tonumber(xaxiscmd.value)]
		wStr(db.major, 'AccelX', db.accelx)
		db.accely = joycmdlist3[tonumber(yaxiscmd.value)]
		wStr(db.major, 'AccelY', db.accely)
		db:set_accel_commands()
		self.active = 'NO'
	end}
	local accel_tab = iup.vbox{
		iup.fill{size = 5},
		iup.label{title = 'Tilt Configuration', font = fontsize},
		custom_accel,
		iup.hbox{
			iup.label{title = 'Left/Right', font = fontsize},
			iup.fill{},
			xaxiscmd,
			iup.fill{expand = 'YES'},
		},
		iup.hbox{
			iup.label{title = 'Fore/Back', font = fontsize},
			iup.fill{},
			yaxiscmd,
			iup.fill{expand = 'YES'},
		},
		iup.fill{size = 15}, 
		iup.hbox{iup.fill{}, set_tilt, iup.fill{}},
		iup.fill{},
		gap = 10,
		tabtitle = 'Accelerometer',
	}

	local tabs = iup.pda_root_tabs{
		buttons_tab,
		accel_tab,
	}

	local main = iup.pdarootframe{
		iup.pdarootframebg{
			iup.vbox{
				--iup.hbox{iup.fill{}, iup.label{title = 'Buttons', size = 200, font = fontsize}, iup.fill{}},
				tabs,
				iup.hbox{iup.fill{}, self.optionsbutton, self.closebutton, iup.fill{}, gap = 15},
				gap=10,
				margin = margin_vbox
			},
		},
	}

	ui.dialog = iup.dialog{
		iup.vbox{
			iup.fill{},
			iup.hbox{
				iup.fill{},
				main,
				iup.fill{},
			},
			iup.fill{},
		},
		defaultesc=self.closebutton,
		bgcolor="0 0 0 *",
		-- fullscreen="YES",
		title = 'DroidButtons v'..db.version..': Button Management',
		border="NO",
		resize="NO",
		menubox="YES",
		topmost="YES",
	}

	function ui.dialog:show_cb()
		tabs:SetTab(2)
		dirtyrecord = false
		set_tilt.active = 'NO'
		custom_accel.value = bool_map[db.use_custom_accel]
		local tmp = joycmdlist2[(db.accelx)]
		xaxiscmd.value = tmp
		xaxiscmd.cursel = tmp
		tmp = joycmdlist2[(db.accely)]
		yaxiscmd.value = tmp
		yaxiscmd.cursel = tmp
	end

	ui.dialog:map()
end
